*
*       Queue manipulation routines to maintain a FIFO list of elements.  The
*       list is singly-linked, with a head and tail pointer.    An arbitrary number
*       of independent queues can be created.  Each list element has a value, V,
*       and a pointer to the next list element.  The value may be any SNOBOL4
*       or program defined data type.
*
*       This package contains the following operations:
*
*   QUEUE
*   APPEND
*   APPENDQQ
*   REMOVE
*   STKQUE
*   QUESTK
*   COPYL
*   PRTQUE
*   CHECKQUE
*   DELEQUE
*   QUENO
*   QUEARY
*   ARYQUE
*   QINIT
*   NORMQUE
*   TBLQUE
*   QUETBL
*   SORTQUE
*
*       To create a new queue, use the data type creation function QUEUE.
*       For example, to create two new queues, Q1 and Q2:
*
*               Q1      =       QUEUE()
*               Q2      =       QUEUE()
*
*       To add a value V to the end of a queue Q, use:
*
*               qAPPEND(V, Q)
*
*       To remove a value from the beginning of queue Q, use:
*
*               V       =       REMOVE(Q)
*
*       The REMOVE function fails if the queue is empty.
*
*       The APPENDQQ function returns a new queue, consisting
*       of the elements of two queues passed as arguments
*
*       The STKQUE function converts a stack to a queue.
*       See package STACK.SNO for the stack functions.
*
*       The QUESTK function converts a queue to a stack.
*
*       COPYL creates a copy of queue Q.
*
*       PRTQUE prints the contents of a queue.
*
*       QUENO returns the number of elements in Q
*
*       QUEARY returns an array from queue Q
*
*       ARYQUE returns a queue from an array
*
*       CHECKQUE checks if an item belongs to a queue.
*
*       DELEQUE deletes an arbitrary element from a queue.
*
*       QINIT discards a queue's contents and reinitializes the queue.
*
*       NORMQUE returns the queue passed as argument "normalized", i.e.
*       with multiple entries reduced to single ones
*
*       QUETBL returns a tanble from a queue
*
*       TBLQUE returns a queue from a table
*
*       SORTQUE sorts a queue of alphabeticized numeric values
*
*
*       From STRING AND LIST PROCESSING IN SNOBOL4 by Ralph E. Griswold,
*       Additions and corrections by Burkhard Meiner (1989)
*
*
        DATA("QUEUE(HEAD,TAIL)")
        DATA("ELEMENT(VALUE,NEXT)")
*       Add an element to the end of a queue.
*
        DEFINE("qAPPEND(V,Q)E")                                  :(qAPPEND_END)
qAPPEND  E       =       TAIL(Q)
        TAIL(Q) =       ELEMENT(V)
        HEAD(Q) =       IDENT(E) TAIL(Q)                        :S(RETURN)
        NEXT(E) =       TAIL(Q)                                 :(RETURN)
qAPPEND_END
*
* Merge two queues: APPEND Q1 TO Q2
*
        DEFINE('APPENDQQ(Q1,Q2)')                               :(APPENDQQ_END)
APPENDQQ
+       APPENDQQ = Q2
        DIFFER(HEAD(APPENDQQ))                                  :F(APPENDQQ1)
        NEXT(TAIL(APPENDQQ)) = HEAD(Q1)
        TAIL(APPENDQQ) = TAIL(Q1)                               :(RETURN)
APPENDQQ1
+       HEAD(APPENDQQ) = HEAD(Q1)
        TAIL(APPENDQQ) = TAIL(Q1)                               :(RETURN)
APPENDQQ_END
*       Remove and return element from beginning of queue.
*
        DEFINE("REMOVE(Q)")                                     :(REMOVE_END)
REMOVE  REMOVE  =       DIFFER(HEAD(Q)) VALUE(HEAD(Q))          :F(FRETURN)
        HEAD(Q) =       NEXT(HEAD(Q))
        TAIL(Q) =       IDENT(HEAD(Q))                          :(RETURN)
REMOVE_END
*       Convert a stack to a queue.  The LIFO nature of the stack is maintained.
*       The argument Q is returned as the value of the function.
*
*       Call: STKQUE(stack name, queue name)
*
        DEFINE("STKQUE(S,STKQUE)")                              :(STKQUE_END)
STKQUE  qAPPEND(POP(S),STKQUE)                                   :S(STKQUE)F(RETURN)
STKQUE_END
*       Convert a queue to a stack.  The FIFO nature of the queue is
*       reversed, and becomes LIFO on the stack.  The argument stack is
*       returned as the value of the function.
*
*       Call: QUESTK(queue name, stack name)
*
*       Note that QUESTK(STKQUE(S,Q),S) reverses the order of elements on stack S,
*       and STKQUE(QUESTK(Q,S),Q) reverses the order of elements on queue Q.
*
        DEFINE("QUESTK(Q,QUESTK)")                              :(QUESTK_END)
QUESTK  PUSH(REMOVE(Q),QUESTK)                                  :S(QUESTK)F(RETURN)
QUESTK_END
*       Create a unique copy of a queue.
*
        DEFINE("COPYL(Q)")                                      :(COPYL_END)
COPYL   COPYL   =       QUEUE()
        E       =       DIFFER(HEAD(Q)) COPY(HEAD(Q))           :F(RETURN)
        HEAD(COPYL)     =       E
COPYLP  NEXT(E) =       DIFFER(NEXT(E)) COPY(NEXT(E))           :F(COPYL1)
        E       =       NEXT(E)                                 :(COPYLP)
COPYL1  TAIL(COPYL)     =       E                               :(RETURN)
COPYL_END
*       Print contents of queue Q:
*
        DEFINE("PRTQUE(Q)P")                                    :(PRTQUE_END)
PRTQUE  P       =       HEAD(Q)
PRTQ1   OUTPUT  =       DIFFER(P) VALUE(P)                      :F(RETURN)
        P       =       NEXT(P)                                 :(PRTQ1)
PRTQUE_END
*
* CHECKQUE checks, if VAL occurs in the queue Q.
* If so, it returns the number of VAL in Q
*
* This function is useful to test, if a certain element belongs
* to a list of items, without removing it from the list itself.
*
        DEFINE("CHECKQUE(VAL,Q)P")                              :(CHECKQUE_END)
CHECKQUE        P = HEAD(Q)
CHECKQ1 CHECKQUE = DIFFER(P) CHECKQUE + 1                       :F(FRETURN)
        IDENT(VALUE(P),VAL)                                     :S(RETURN)
        P = NEXT(P)                                             :(CHECKQ1)
CHECKQUE_END
*
* Function to remove an arbitrary item from a queue:
*
        DEFINE('DELEQUE(VAL,Q)P,R')                             :(DELEQUE_END)
DELEQUE P = HEAD(Q); DIFFER(P)                                  :F(FRETURN)
DELEQUE1        IDENT(VALUE(P),VAL)                             :S(DELEQUE2)
        Q = P
        P = NEXT(P)
        DIFFER(P)                                       :S(DELEQUE1)F(FRETURN)
DELEQUE2
+       NEXT(Q) = NEXT(P)                                       :(RETURN)
DELEQUE_END
*
* QUENO Returns number of elements in queue
*
        DEFINE('QUENO(Q)P')                                     :(QUENO_END)
QUENO   P = HEAD(Q)
        DIFFER(P)                                               :F(FRETURN)
QUENO_L QUENO = QUENO + 1
        P = NEXT(P)
        DIFFER(P)                                       :S(QUENO_L)F(RETURN)
QUENO_END
*
* QUEARY returns an array from queue Q
*
        DEFINE('QUEARY(Q)P,I')                          :(QUEARY_END)
QUEARY  QUEARY = QUENO(Q)
        QUEARY = NE(QUEARY) ARRAY(CONVERT(QUEARY,'STRING'))     :F(FRETURN)
        P = HEAD(Q)
QUE_L   I = I + 1
        QUEARY<I> = VALUE(P)                            :F(RETURN)
        P = NEXT(P)                                     :(QUE_L)
QUEARY_END
        DEFINE('ARYQUE(ARY)I,J')                         :(ARYQUE_END)
ARYQUE  ARYQUE = QUEUE()
ARYQUE1 I = I + 1
        J = ARY<I>                                      :F(RETURN)
        qAPPEND(J,ARYQUE)                                :(ARYQUE1)
ARYQUE_END
*       Discard a queue's contents and reinitialize the queue:
*
        DEFINE("QINIT(Q)")                                      :(QINIT_END)
QINIT   HEAD(Q) =
        TAIL(Q) =                                               :(RETURN)
QINIT_END
*
*       Normalize a queue: Return a queue with each separate item
*       occuring only once
*
        DEFINE('NORMQUE(QUEUE)P,I,X')                           :(NORMQUE_END)
NORMQUE P = HEAD(QUEUE)                                         :F(FRETURN)
        DIFFER(P)                                               :F(FRETURN)
        NORMQUE = TABLE()
NORMQUE1
+       I = I + 1
        NORMQUE<VALUE(P)> = 1
        P = NEXT(P)
        DIFFER(P)                                               :S(NORMQUE1)
        NORMQUE = CONVERT(NORMQUE,'ARRAY')
        I = 0
        X = QUEUE()
NORMQUE2
+       I = I + 1
        P = NORMQUE<I,1>                                        :F(NORMQUE3)
        qAPPEND(P,X)                                             :(NORMQUE2)
NORMQUE3
+       NORMQUE = X                                             :(RETURN)
NORMQUE_END
*
* QUETBL
*
        DEFINE('QUETBL(QUEUE)P,R')                              :(QUETBL_END)
QUETBL  QUETBL = TABLE()
        P = HEAD(QUEUE)
QUETBL1 R = VALUE(P)
        QUETBL<R> = 1
        P = NEXT(P)
        DIFFER(P)                                       :S(QUETBL1)F(RETURN)
QUETBL_END
        DEFINE('TBLQUE(TABLE)P,I')                              :(TBLQUE_END)
TBLQUE  TBLQUE = QUEUE()
        TABLE = CONVERT(TABLE,'ARRAY')                          :F(FRETURN)
TBLQUE1 I = I + 1
        P = TABLE<I,1>                                          :F(RETURN)
        qAPPEND(P,TBLQUE)                                        :(TBLQUE1)
TBLQUE_END
+       DEFINE('SORTQUE(Q)I,A,B,C')                             :(SORTQUE_END)
SORTQUE
+       SORTQUE = QUEARY(Q)                                     :F(FRETURN)
        I = 0
*
* First, pad entries
*
SORTQUE_PAD_L
+       I = I + 1
        B = 0
        A = SORTQUE<I>                          :F(SORTQUE_PAD_OK)
        A ? POS(0) LEN(8) $ C =                 :S(SORTQUE_PAD_GO)
        ERROR(6)                                :(FRETURN)
SORTQUE_PAD_GO
+       C = REVERSE(C)
        C POS(0) SPAN(CHAR(0)) =
        C = REVERSE(C)
        C = TRIM(C)
        C ? POS(0) SPAN(' 0') =
        C ? POS(0) SPAN(NUMBERS) @B
        C = DUPL(' ',8 - B) C
        C = RPAD(C,16)
        A = C A
        SORTQUE<I> = A                          :(SORTQUE_PAD_L)
SORTQUE_PAD_OK
+       SORTQUE = SORT(SORTQUE)
*
* restore entries:
*
        I = 0
SORTQUE_DEPAD_L
+       I = I + 1
        A = SORTQUE<I>                          :F(SORTQUE_DEPAD_OK)
        A POS(0) LEN(16) $ C =
        C = TRIM(C)
        C POS(0) SPAN(' ') =
        C = RPAD(C,8,CHAR(0))
        A = C A
        SORTQUE<I> = A                          :(SORTQUE_DEPAD_L)
SORTQUE_DEPAD_OK
+        SORTQUE = ARYQUE(SORTQUE)              :(RETURN)
SORTQUE_END
